<?php

/**
 * @file
 * State Flow implementation of the State Machine class.
 */

class StateFlowEntity extends StateMachine {
  /**
   * Called from StateMachine::__construct to initialize the states and events.
   */
  public $initial;
  public $current;
  public $default_state;
  public $object;
  public $plugin;

  public function __construct($plugin = NULL) {
    $this->plugin = $plugin;
    $this->set_object($plugin['object']);
    $this->init($plugin['handler']['workflow_options']);
    $this->set_current_state((string) $this->load());
  }

  /**
   * Called from StateMachine::__construct to initialize the states and events.
   *
   * Define two states.
   * First revision:
   *  - Expose go to draft button
   *  - Expose go to publish button
   *  - Upon publish, create new revision (handled outside of state machine)
   * Second revision:
   *  - Menu alter edit link to load most recent revision (whether published or
   *      revision from states).
   *  - On hook_nodeapi (op: load), force new revision checkbox on node edit
   *      form.
   *    - Expose go to draft button
   *  - Create new revision, prevent node table from updating new revision as
   *       published revision.
   *  - Expose go to publish button
   *  - Upon publish, set revision id in node table
   *  - Repeat as needed
   */
  public function init($workflow = array()) {
    // Initialize states.
    foreach ($workflow['states'] as $name => $options) {
      // Some methods require $this when called by State Machine e.g. on_exit().
      foreach ($options as $key => $value) {
        if ($key == 'on_exit' || $key == 'on_enter') {
          $options[$key] = array($this, $value);
        }
      }
      $this->create_state($name, $options);
    }

    // Initialize events.
    foreach ($workflow['events'] as $name => $options) {
      $this->create_event($name, $options);
    }

    $this->default_state = isset($workflow['default_state']) ? $workflow['default_state'] : NULL;

    // Find publishable vid.
    $this->set_available_publish_revision();

    // Set latest history.
    if (!empty($this->object) && isset($this->object->vid)) {
      $this->set_current_history();
    }
  }

  /**
   * Set the vid from the node table as the "Published" revision.
   *
   * If the node is not currently published then there is no Published revision.
   * It only needs to set this id when the entity first loads.  Keep that test
   * to see if it's already set or it gets confused when creating new
   * revisions.
   */
  public function set_available_publish_revision() {
    if (!isset($this->object->published_revision_id) && !empty($this->object->{$this->get_revision_key()})) {
      $this->object->published_revision_id = 0;
      $entity_type = $this->get_entity_type();
      $entity_info = entity_get_info($entity_type);
      if (isset($entity_info['base table'])) {
        $published_revision = db_select($entity_info['base table'])
          ->fields('node', array($this->get_revision_key()))
          ->condition('nid', $this->object->{$this->get_entity_id_key()})
          ->condition('status', 1)
          ->execute()->fetchCol();
        if (!empty($published_revision)) {
          $this->object->published_revision_id = $published_revision[0];
        }
      }
    }
  }

  /**
   * Provide history with each revision.
   *
   * @todo, this does not need to be in state_flow_node.inc. It could easily get
   * abstracted to state_flow_entity.inc
   */
  public function set_current_history() {
    // Set defaults.
    $this->object->current_state = '';
    $this->object->current_log = '';
    $this->object->current_state_timestamp = '';

    $result = db_select('state_flow_history', 'sfh')
      ->fields('sfh', array('state', 'timestamp', 'log'))
      ->condition('entity_id', $this->object->{$this->get_entity_id_key()})
      ->condition('revision_id', $this->object->{$this->get_revision_key()})
      ->orderBy('timestamp', 'DESC')
      ->range(0, 1)
      ->execute()
      ->fetchObject();
    // Set actual data if found any.
    if ($result) {
      $this->object->current_state = $result->state;
      $this->object->current_log = $result->log;
      $this->object->current_state_timestamp = $result->timestamp;
    }
  }

  /**
   * Return an event.
   *
   * @param string $key
   *   The key of the event to return.
   *
   * @return bool|StateMachine_Event
   *   The event object. Or FALSE if the event doesn't exist.
   */
  public function get_event($key) {
    if (is_null($key) || !array_key_exists($key, $this->events)) {
      return FALSE;
    }

    if (is_array($this->events[$key])) {
      $options = $this->events[$key];
      $this->events[$key] = new StateFlow_Event($key, $this, $options);
    }

    return $this->events[$key];
  }

  public function get_object() {
    return $this->object;
  }

  public function set_object($object) {
    $this->object = $object;
    // Set a reference to this machine to the object.
    $object->state_flow = $this;
  }

  public function get_states_options() {
    return $this->states;
  }

  /**
   * Returns TRUE if the entity is new.
   *
   * @return bool
   *   TRUE if the entity is new.
   */
  public function object_is_new() {
    // If the object is available, and has and id and isn't marked as new it's
    // an existing one - otherwise it's new.
    if (($object = $this->get_object()) && !empty($object->{$this->get_entity_id_key()}) && empty($object->is_new)) {
      return FALSE;
    }
    return TRUE;
  }

  public function object_is_drafty_forward_revision() {
    return (($entity = $this->get_object()) && !empty($entity->revision) && !empty($entity->is_new_revision) && !empty($entity->default_revision));
  }

  /**
   * Get the label for the current state.
   *
   * @return string
   *   Label for the current state.
   */
  public function get_label_for_current_state() {
    $state_key = $this->get_current_state();
    $state = $this->get_state($state_key);
    if ($label = $state->get_option('label')) {
      return $label;
    }
    else {
      return drupal_ucfirst($state_key);
    }
  }

  public function get_entity_type() {
    if (isset($this->plugin['handler']['entity_type'])) {
      return $this->plugin['handler']['entity_type'];
    }
  }

  /**
   * Get the revision key for this entity type.
   */
  public function get_revision_key() {
    $entity_type = $this->get_entity_type();
    $entity_info = entity_get_info($entity_type);
    return $entity_info['entity keys']['revision'];
  }

  /**
   * Get the revision key for this entity type.
   */
  public function get_entity_id_key() {
    $entity_type = $this->get_entity_type();
    $entity_info = entity_get_info($entity_type);
    return $entity_info['entity keys']['id'];
  }

  /**
   * Get the state flow history entity.
   */
  public function get_history_entity() {
    $entity_type = $this->get_entity_type();
    $id_key = $this->get_entity_id_key();
    $revision_key = $this->get_revision_key();

    // Check if the entity has a prepared history entity attached.
    $entity = $this->get_object();
    if (!empty($entity->state_flow_history_entity)) {
      $this->history_entity = $entity->state_flow_history_entity;
    }
    if (empty($this->history_entity)) {
      // Make a new empty class for the history entity.
      $history_entity = new stdClass();

      // If this object has a revision key or id key, get it's previous history.
      // If this is a new object, like on the node add form, it won't have a
      // previous history object.
      if (!empty($entity->$revision_key) || !empty($entity->$id_key)) {

        $query = db_select('state_flow_history', 'sfh')
                  ->fields('sfh', array('hid'))
                  ->orderBy('hid', 'DESC')
                  ->range(0, 1)
                  ->condition('entity_type', $entity_type);

        // Set a condition for the vid or nid.
        if (!empty($entity->$revision_key)) {
          $query->condition('revision_id', $entity->$revision_key);
        }
        elseif (!empty($entity->$id_key)) {
          $query->condition('entity_id', $entity->$id_key);
        }

        // Get the results of the query.
        $result = $query->execute()->fetchCol('hid');

        // Load up the fields of the history entity.
        if (!empty($result[0])) {
          $hid = $result[0];
          $history_entities = entity_load('state_flow_history_entity', array('hid' => $hid));
          if (!empty($history_entities[$hid])) {
            $history_entity = clone $history_entities[$hid];
          }
        }

        // Unset the hid and the log message so those don't carry over.
        unset($history_entity->hid);
        unset($history_entity->log);
      }
      $this->history_entity = $history_entity;
      $this->history_entity->state = $this->get_current_state();
    }

    // Populate the entity with current information.
    // Not sure how much of this is necessary.
    $this->history_entity->revision_id = NULL;
    if (!empty($entity->$revision_key)) {
      $this->history_entity->revision_id = $entity->$revision_key;
    }

    $this->history_entity->entity_id = NULL;
    if (!empty($entity->$id_key)) {
      $this->history_entity->entity_id = $entity->$id_key;
    }

    $this->history_entity->timestamp = REQUEST_TIME;
    $this->history_entity->entity_type = $entity_type;
    return $this->history_entity;
  }

  /**
   * Get available events as an array suitable for FAPI.
   *
   * Entities that aren't the published active revision can't transition to
   * unpublished since this would take down the published revision too.
   *
   * @param bool $validate_options
   *   A boolean for whether events should be validated. For instance, a node in
   *   'Draft' state might have 'Needs Review' and 'Published' as available next
   *   states/events. However the currently logged in user might not have
   *   permission to perform the publishing event. Passing TRUE to
   *   $validate_events would exclude the 'Published' option for that user.
   *
   * @return array
   *   Array of events. Keyed by the machine name, value is the label.
   */
  public function get_available_events_options($validate_options = TRUE) {
    $events = $this->get_available_events();
    $return = array();
    $current_state = $this->get_current_state();
    /** @var  StateFlow_Event $event */
    foreach ($events as $event_machine_name => $event) {
      if (empty($validate_options) || $event->validate()) {
        // Only the active published revision can transition to a state that is
        // unpublished.
        $target_state = $event->get_target_state();
        if (!$target_state->is_unpublished() || $this->isActivePublishedRevision() || $target_state->name == $current_state) {
          $return[$event_machine_name] = $event->get_option('label');
        }
      }
    }
    return $return;
  }

  /**
   * Set the state flow history entity.
   */
  public function set_history_entity($history_entity) {
    $this->history_entity = $this->clean_history_entity($history_entity);
  }

  /**
   * Removes all non Entity API properties from the history_entity.
   */
  public function clean_history_entity($history_entity) {

    // If entity module is available, uses it's function to get all the
    // properties and fields on the state_flow_history_entity.
    if (module_exists('entity')) {
      $sfhe_properties = array_keys(entity_get_all_property_info('state_flow_history_entity'));
    }
    // Otherwise get the properties and fields directly.
    else {
      $property_info = state_flow_entity_entity_property_info();
      $fields = field_info_instances('state_flow_history_entity', 'state_flow_history_entity');
      $sfhe_properties = array_merge(array_keys($fields), array_keys($property_info['state_flow_history_entity']['properties']));
    }

    // @todo, These should be in the properties already.
    $sfhe_properties = array_merge(array('event', 'event_comment'), $sfhe_properties);

    $clean_history_entity = new stdClass();
    foreach ($sfhe_properties as $property) {
      if (!empty($history_entity->{$property})) {
        $clean_history_entity->{$property} = $history_entity->{$property};
      }
    }

    return $clean_history_entity;
  }

  /**
   * Validate the filed API from values with state_flow_history_entity values.
   */
  public function history_entity_form_field_validate($form, &$form_state) {
    entity_form_field_validate('state_flow_history_entity', $form, $form_state);
  }

  /**
   * Build up a full state flow history entity from form values.
   *
   * Called from state_flow_events_revision_submit().
   */
  public function history_entity_form_submit_build_entity($form, &$form_state) {
    $history_entity = $this->get_history_entity();
    entity_form_submit_build_entity('state_flow_history_entity', $history_entity, $form, $form_state);
    $this->set_history_entity($history_entity);
  }

  /**
   * Provide entire state_flow history.
   *
   * @todo, this method should use dbtng and specify entity_type.
   */
  public function get_full_history() {
    $history = db_query('
    SELECT sfh.*, u.uid, u.name AS user_name
    FROM {state_flow_history} sfh
    LEFT JOIN {users} u ON u.uid = sfh.uid
    WHERE sfh.entity_id = :entity_id
    ORDER BY sfh.timestamp DESC',
          // Add entity_type where.
    array(
      ':entity_id' => $this->object->entity_id,
    ))->fetchAll();

    return $history;
  }

  /**
   * Check if the current object is a draft revision.
   *
   * A draft revision is it if:
   * - It's already marked as draft revision.
   * - It's not a new entity.
   * - The currently active revision differs from the entity revision.
   * - There is an active revision and a new revision is requested.
   * - A requested revision is not published.
   *
   * Also sets the property is_draft_revision on the entity.
   *
   * @return bool
   *   Whether or not the current object is a draft revision.
   */
  protected function isDraftRevision() {
    $entity = $this->get_object();
    // If this somehow was marked as is_draft_revision we don't change this.
    if (empty($entity->is_draft_revision)) {
      // Ensure this has a value - empty() also passes if the property wasn't
      // set yet.
      $entity->is_draft_revision = FALSE;
      // If this revision isn't the published one it's likely a draft.
      $this->set_available_publish_revision();

      if (!$this->object_is_new() && !empty($entity->published_revision_id)) {
        // If the target state is neither published nor unpublished, this is a draft.
        $entity->is_draft_revision = (empty($entity->event) || !($event = $this->get_event($entity->event)) || (!$event->get_target_state()->is_published() && !$event->get_target_state()->is_unpublished()));
      }
    }
    return $entity->is_draft_revision;
  }

  /**
   * Check if the current object is the active published revision.
   *
   * A active published revision is it if:
   * - The object isn't ignored.
   * - The published_revision_id property is empty.
   * - The published_revision_id property is equal to the current revision id.
   *
   * @return bool
   *   Whether or not the current object is a the active published revision.
   */
  public function isActivePublishedRevision() {
    $entity = $this->get_object();
    if (!$this->ignore()) {
      $this->set_available_publish_revision();
      return (empty($entity->published_revision_id) || $this->object->{$this->get_revision_key()} == $entity->published_revision_id);
    }
    return !empty($entity->status);
  }

  /**
   * Called by hook_entity_presave().
   *
   * @return bool
   *   TRUE if successful, FALSE otherwise.
   *
   * @see state_flow_entity_entity_presave()
   */
  public function entityPresave() {
    $entity = $this->get_object();

    // Set the target status if it can be determined.
    if (!empty($entity->event)) {
      $event = $this->get_event($entity->event);
      $state = $event->get_target_state();
      if ($state) {
        $entity->status = ($state->is_published()) ? NODE_PUBLISHED : NODE_NOT_PUBLISHED;
      }
    }
    elseif ($this->object_is_new()) {
      // If this is a new entity without an event preset the default event. This
      // ensures programmatic saves work.
      $entity->event = $this->get_default_event_name();
    }
    else {
      // @TODO Do we need a special handling in such cases?
      $entity->event = $this->get_default_event_name();
    }
    $event = $entity->event;

    // Check if this is a draft revision.
    $entity->is_draft_revision = $this->isDraftRevision();

    if (($event = $this->get_event($event)) && ($new_state = $event->execute())) {
      $history_entity = $this->get_history_entity();
      $history_entity->event = $event->name;
      $this->set_history_entity($history_entity);

      // Allow the previous state to run its 'on_exit' callbacks.
      $this->get_state($this->get_current_state())->on_exit();

      // Set the new state.
      $this->set_current_state($new_state);

      // Allow the new state to run its 'on_enter' callbacks.
      $this->get_state($this->get_current_state())->on_enter();
      return TRUE;
    }
    $this->on_event_fail($event);
    return FALSE;
  }

  /**
   * Extending fire_event() from state_machine's base.inc.
   *
   * Extending fire_event() from state_machine's base.inc to add uid and log
   * arguments.
   *
   * @param string $key
   *   The event name.
   * @param int $uid
   *   The user id this event is fired for.
   * @param string $log
   *   The log message for the history.
   *
   * @return bool
   *   TRUE if successful, FALSE otherwise.
   *
   * @see state_flow_entity_entity_presave()
   */
  public function fire_event($key, $uid = NULL, $log = '') {
    if (($event = $this->get_event($key))) {
      $object = $this->get_object();
      // Set the event properties on the entity.
      $object->event = $key;
      $object->event_comment = $log;
      $object->event_uid = $uid;
      // Save the object, all subsequent handling is done using the entity API
      // hooks and drafty.
      entity_save($this->get_entity_type(), $object);
      return TRUE;
    }
    $this->on_event_fail($event);
    return FALSE;
  }

  /**
   * Called by hook_entity_insert() / hook_entity_update().
   *
   * @param string $key
   *   The event name.
   * @param int $uid
   *   The user id this event is fired for.
   * @param string $log
   *   The log message for the history.
   *
   * @return bool
   *   TRUE if successful, FALSE otherwise.
   *
   * @see _state_flow_entity_entity_update()
   */
  public function entity_saved($key, $uid = NULL, $log = '') {
    $entity = $this->get_object();
    $is_drafty_forward_revision = (!empty($entity->revision) && !empty($entity->is_new_revision) && !empty($entity->default_revision));

    if (!$is_drafty_forward_revision && ($event = $this->get_event($key))) {
      // Ensure the proper id's are set.
      $history_entity = $this->get_history_entity();
      $history_entity->entity_id = $entity->{$this->get_entity_id_key()};
      $history_entity->revision_id = $entity->{$this->get_revision_key()};
      $history_entity->event = $key;
      // Ensure we create a new history entry for new revisions.
      if (!empty($entity->revision)) {
        unset($history_entity->hid);
      }
      $this->set_history_entity($history_entity);

      // Write a history record for this state change.
      $this->write_history($uid, $log);

      // Set this revision as active and update the current status in the table.
      $this->write_active();

      // Allow the event to "finish".
      $event->finish();

      // Allow state_flow to provide other hooks or event triggers.
      //
      // @todo the original hook in 7.x-2.x looks like this:
      //   state_flow_invoke_event_handlers($this, $key, $uid, $log);
      // We should review to see if other code uses this hook and update as
      // needed.
      state_flow_entity_invoke_event_handlers($entity, $this->get_current_state(), $this->get_history_entity());
    }
    elseif ($is_drafty_forward_revision) {
      // Forward the history entities.
      $this->forward_history($entity->state_machine_loaded_revision, $entity->{$this->get_revision_key()});
    }
    else {
      $this->on_event_fail($event);
      return FALSE;
    }
  }

  /**
   * Forwards the history and states similar what drafty does for entities.
   *
   * @param $fromRevision
   * @param $toRevision
   */
  protected function forward_history($from_revision, $to_revision) {
    if ($to_revision) {
      // Fetch the matching history entries.
      $query = db_select('state_flow_history', 'sfh')
        ->fields('sfh', array('hid'))
        ->orderBy('hid', 'ASC')
        ->condition('entity_type', $this->get_entity_type())
        ->condition('revision_id', $from_revision)
        ->condition('entity_id', $this->get_object()->{$this->get_entity_id_key()})
        ->execute()
        ->fetchAllKeyed(0, 0);
      if (!empty($query)) {
        // We don't manipulate this in the main table because this are entities
        // and as such modules could use hook to do stuff. Otherwise we might
        // also risk the data consistency of fields.
        $history_entities = entity_load('state_flow_history_entity', $query);
        foreach ($history_entities as $i => $history_entity) {
          // Fetch related state.
          $state = db_select('state_flow_states', 'sfs')
            ->fields('sfs')
            ->range(0, 1)
            ->condition('hid', $history_entity->hid)
            ->condition('revision_id', $from_revision)
            ->condition('entity_type', $history_entity->entity_type)
            ->execute()
            ->fetchAssoc();

          // Unset hid - to ensure a new entity is created.
          unset($history_entity->hid);
          $history_entity->revision_id = $to_revision;

          // @TODO Build proper entity handling so that we can use
          // entity_save().
          field_attach_presave('state_flow_history_entity', $history_entity);
          // Write into the state_flow_history table.
          drupal_write_record('state_flow_history', $history_entity);
          // Update field tables.
          field_attach_update('state_flow_history_entity', $history_entity);

          // Insert state for new history entity.
          if ($state) {
            unset($state['state_flow_id']);
            $state['hid'] = $history_entity->hid;
            $state['revision_id'] = $history_entity->revision_id;
            db_insert('state_flow_states')
              ->fields($state)
              ->execute();
          }
        }
      }
    }
  }

  /**
   * Write the transaction to the history table.
   *
   * @param int $uid
   *   The user id.
   * @param string $log
   *   The message to log.
   *
   * @return object
   *   The history entity object.
   */
  public function write_history($uid, $log = '') {
    // Start with what the db has.
    $history_entity = $this->get_history_entity();

    // @todo, this was added to accommodate the double saving that happens with
    // node forms.
    // Reevaluate if necessary.
    if (empty($log) && !empty($history_entity->event_comment)) {
      $log = $history_entity->event_comment;
    }

    $entity_type = $this->get_entity_type();
    $entity_info = entity_get_info($entity_type);
    $id_key = $entity_info['entity keys']['id'];
    $revision_key = $entity_info['entity keys']['revision'];

    // Apply current settings.
    if (!empty($this->object->$revision_key)) {
      $history_entity->revision_id = $this->object->$revision_key;
    }
    else {
      $history_entity->revision_id = NULL;
    }
    if (!empty($this->object->$id_key)) {
      $history_entity->entity_id = $this->object->$id_key;
    }
    else {
      $history_entity->entity_id = NULL;
    }

    $history_entity->from_state = $history_entity->state;
    $history_entity->state = $this->get_current_state();
    $history_entity->timestamp = REQUEST_TIME;
    $history_entity->uid = $uid;
    $history_entity->log = $log;
    if (!empty($state)) {
      $history_entity->state = $state;
    }
    $history_entity->entity_type = $entity_type;

    // Often we are writing a new history entity and this property doesn't
    // exist.
    // The way node form submitting requires writing a history on form submit
    // and then updating it in entity_update when the new vid is available. That
    // means a hid might be present.
    if (empty($history_entity->hid)) {
      $primary_keys = array();
    }
    else {
      $primary_keys = array('hid');
    }

    // @todo, tests needed here.
    field_attach_presave('state_flow_history_entity', $history_entity);

    // Update the state_flow_history table.
    drupal_write_record('state_flow_history', $history_entity, $primary_keys);
    // Update field tables.
    field_attach_update('state_flow_history_entity', $history_entity);

    $this->set_history_entity($history_entity);
    return $history_entity;
  }

  /**
   * Get the active revision.
   *
   * The active revision is not necessarily the published revision.
   *
   * @todo I am not sure I like this method. confirm that we need this. There
   * may be a better way.
   *
   * @return int
   *   The revision_id.
   */
  public function get_active_revision() {
    if (empty($this->object->active_revision_id)) {
      $entity_type = $this->get_entity_type();
      $entity_info = entity_get_info($entity_type);
      $id_key = $entity_info['entity keys']['id'];
      $this->object->active_revision_id = 0;

      if (!$this->object_is_new()) {
        $revision_id = db_select('state_flow_states', 'sfs')
          ->fields('sfs', array('revision_id'))
          ->condition('entity_type', $entity_type)
          ->condition('entity_id', $this->object->$id_key)
          ->condition('active', 1)
          ->execute()
          ->fetchCol();
        if (!empty($revision_id[0])) {
          $this->object->active_revision_id = $revision_id[0];
        }
      }
    }
    return $this->object->active_revision_id;
  }

  /**
   * Get the latest revision id.
   *
   * @param int $entity_id
   *   The entity id of the entity to fetch the latest revision id for.
   *   DEPRECATED!
   *
   * @return int|FALSE
   *   The latest revision id or FALSE on failure.
   */
  public function get_latest_revision($entity_id = NULL) {
    $entity_type = $this->get_entity_type();
    $entity_info = entity_get_info($entity_type);
    if (isset($entity_info['revision table'])) {
      $id_key = $entity_info['entity keys']['id'];
      $revision_key = $entity_info['entity keys']['revision'];
      if (empty($entity_id)) {
        $entity_id = $this->object->{$id_key};
      }
      $query = db_select($entity_info['revision table'])
        ->fields($entity_info['revision table'], array($revision_key))
        ->condition($id_key, $entity_id)
        ->orderBy('timestamp', 'DESC')
        ->range(0, 1);
      return $query->execute()->fetchField();
    }
    return FALSE;
  }

  /**
   * Update the current status record for the revision.
   *
   * Only one revision may be active at a time. If additional revisions need to
   * be active, please use fields for now.
   *
   * Using $this->history_entity assumes that write_history() has already
   * been called.
   *
   * called from fire_event() and update_entity_state() when the user is
   * submitting the entity's edit form (like node edit for example).  When
   * called twice we must match the hid but if it's straight from the
   * state_flow_entity transition form state_flow_entity_events_revision() we
   * know that the revision exists and is not new
   *
   * @see state_flow_entity_events_revision()
   */
  public function write_active() {
    $this->write_state($this->history_entity, TRUE);
  }

  /**
   * Update the current status record for the revision.
   *
   * Only one revision may be active at a time. If additional revisions need to
   * be active, please use fields for now.
   *
   * @param object $history_entity
   *   The related history entity.
   * @param bool $active
   *   Define if this is the active revision.
   *
   * @see state_flow_entity_events_revision()
   */
  public function write_state($history_entity, $active = TRUE) {
    // First default all revisions to inactive status.
    // @todo, can this happen in a less hard-coded fashion to accommodate sites
    // That may need multiple active revisions?
    if ($active) {
      db_update('state_flow_states')
        ->fields(array(
          'active' => 0,
        ))
        ->condition('entity_id', $history_entity->entity_id)
        ->condition('entity_type', $history_entity->entity_type)
        ->execute();
    }

    // Store current status, add a new record if it's a new revision or a new
    // entity.
    if (isset($this->object->revision) || $this->object_is_new()) {
      if ($this->object->revision || $this->object_is_new()) {
        db_insert('state_flow_states')
          ->fields(array(
            'entity_type' => $history_entity->entity_type,
            'entity_id' => $history_entity->entity_id,
            'revision_id' => $history_entity->revision_id,
            'hid' => $history_entity->hid,
            'state' => $this->get_current_state(),
            'active' => (int) $active,
            'timestamp' => REQUEST_TIME,
          ))
          ->execute();
      }
      else {
        // Set existing revision as active so we can hit this call twice when
        // editing a node. Require hid so that only one record is updated.
        db_update('state_flow_states')
          ->fields(array(
            'active' => (int) $active,
            'state' => $this->get_current_state(),
            'hid' => $history_entity->hid,
          ))
          ->condition('entity_id', $history_entity->entity_id)
          ->condition('entity_type', $history_entity->entity_type)
          ->condition('revision_id', $history_entity->revision_id)
          ->execute();
      }
    }
    else {
      // Not a new revision AND transition only i.e. the user did not come
      // from an entity edit form like node/%node/edit unlike the db_update
      // above, we do not need to require the hid because a transition-only
      // operation only hits this method once.
      db_update('state_flow_states')
        ->fields(array(
          'active' => (int) $active,
          'state' => $this->get_current_state(),
          'hid' => $history_entity->hid,
        ))
        ->condition('entity_id', $history_entity->entity_id)
        ->condition('entity_type', $history_entity->entity_type)
        ->condition('revision_id', $history_entity->revision_id)
        ->execute();
    }
  }

  public function delete_state_flow_revision($revision_id, $entity_type) {
      // @todo Deleting field data needs testing.
    field_attach_delete_revision('state_flow_history_entity', $this->get_history_entity());
    // Delete entity records.
    db_delete('state_flow_history')
      // ->condition('entity_id', $entity_id)
      ->condition('revision_id', $revision_id)
      ->condition('entity_type', $entity_type)
      ->execute();
    db_delete('state_flow_states')
      // ->condition('entity_id', $entity_id)
      ->condition('revision_id', $revision_id)
      ->condition('entity_type', $entity_type)
      ->execute();
  }

  /**
   * Provide the current state of this revision or entity.
   *
   * @return string
   *   The state machine name.
   */
  public function load() {
    $state = FALSE;
    $entity_type = $this->get_entity_type();
    $entity_info = entity_get_info($entity_type);
    $id_key = $entity_info['entity keys']['id'];
    $revision_key = $entity_info['entity keys']['revision'];

    if (!empty($this->object->$id_key) && !empty($this->object->$revision_key)) {
      $state = $this->revision_state($this->object->$id_key, $this->object->$revision_key);
    }
    elseif (!empty($this->object->$id_key)) {
      $state = $this->latest_state($this->object->$id_key);
    }
    return $state;
  }

  /**
  * Return the state of the revision
  */
  public function revision_state($entity_id, $revision_id) {
    $latest_state = db_select('state_flow_history', 'sfh')
      ->fields('sfh', array('state'))
      ->condition('entity_id', $entity_id)
      ->condition('revision_id', $revision_id)
      ->orderBy('hid', 'DESC')
      ->range(0, 1)
      ->execute()
      ->fetchCol('state');

    return !empty($latest_state[0]) ? $latest_state[0] : FALSE;
  }

  public function latest_state($entity_id) {
    $latest_state = db_select('state_flow_history', 'sfh')
      ->fields('sfh', array('state'))
      ->condition('entity_id', $entity_id)
      ->orderBy('hid', 'DESC')
      ->range(0, 1)
      ->execute()
      ->fetchCol('state');

    return !empty($latest_state[0]) ? $latest_state[0] : FALSE;
  }

  /**
   * Get a default event.
   *
   * When the event form is embedded on the node edit form or elsewhere it may
   * be the case that the state does not change. The method checks for an event
   * that has a target state that matches the current state.
   *
   * @return string|NULL
   *   An event name or NULL.
   */
  public function get_default_event_name() {

    foreach ($this->get_available_events() as $event) {

      $target = $event->get_option('target');
      if ($target === $this->current) {
        return $event->name;
      }
    }

    return NULL;
  }


  /**
  * @todo, this may be the best way to do validation on form submission.
    public function validate_event($event_machine_name, $form, &$form_state) {
      $event = $this->get_event($event_machine_name);
      return $event->validate();
    }
  *
  */

  /**
   * Remove scheduled events if the new state isn't scheduled.
   */
  public function on_exit_scheduled() {
    $event = $this->get_event($this->get_history_entity()->event);
    if (module_exists('state_flow_schedule') && $event->get_option('target') !== 'schedule') {
      db_query('
        DELETE FROM state_flow_schedule
        WHERE entity_type = :entity_type
        AND entity_id = :entity_id
        AND revision_id = :revision_id',
        array(
          ':entity_type' => $this->get_entity_type(),
          ':entity_id' => $this->object->{$this->get_entity_id_key()},
          ':revision_id' => $this->object->{$this->get_revision_key()},
        )
      );
    }
  }

}

/**
 * @todo, should this class be renamed to StateFlow_Event?
 */
class StateFlow_Event extends StateMachine_Event {
  public function get_machine() {
    return $this->machine;
  }

  public function validate() {
    if (parent::validate()) {
      if (!empty($this->options['permissions'])) {
        foreach ($this->options['permissions'] as $permission) {
          if (!user_access($permission)) {
            return FALSE;
          }
        }
      }
      return TRUE;
    }
    return FALSE;
  }
}
