/**
 * @file
 */
(function($) {
  /**
   * Periodically ping the notifications server for new notifications.
   */
  Drupal.behaviors.OpenEDUNotificationsClientBadge = {
    attach: function(context, settings) {
      var local_refresh_interval = 5000; // ms
      // Set up vars.
      var loading_class = 'loading',
          endpoint = settings.OpenEDUNotificationsClient.endpoint,
          badge_wrapper = settings.OpenEDUNotificationsClient.badge_wrapper,
          request_interval = settings.OpenEDUNotificationsClient.request_interval;

      var subscriptions_endpoint = settings.OpenEDUNotificationsClient.subscriptions_endpoint,
          subscriptions_badge_wrapper = settings.OpenEDUNotificationsClient.subscriptions_badge_wrapper,
          subscriptions_request_interval = settings.OpenEDUNotificationsClient.request_interval;

      // Check if badgeWrapper is present.
      var $badge_wrapper = $('#' + badge_wrapper);
      var $subscriptions_badge_wrapper = $('#' + subscriptions_badge_wrapper),
          $subscriptionListTrigger = $subscriptions_badge_wrapper.find('a.subscription-list-trigger');

      if ($subscriptionListTrigger.length) {
        // Click.
        $subscriptionListTrigger.click(function(e) {
          e.preventDefault();
          e.stopPropagation();
          $subscriptions_badge_wrapper.find('.list').toggle();
        });

        // Hide when clicking outside.
        $(document).on('click', function(event) {
          var subscriptionList = $subscriptions_badge_wrapper.find('.list');

          if (subscriptionList.is(':visible')) {
            subscriptionList.hide();
          }
        });
      }

      if ($subscriptions_badge_wrapper.length && subscriptions_endpoint) {
        get_subscriptions_badge();
      }

      if ($badge_wrapper.length && endpoint) {
        // Initial notification badge load.
        get_notification_badge();
      }

      /**
       * This will handle the running of notification requests.
       *
       * @param namespace
       *  A name, is the string version of the name of 'loop_callback'.
       * @param pre_callback
       *  Callback for what to have happen before the request is run, only if the request is to be run.
       * @param post_callback
       *  Callback for what to do after we have return data, regardless if request has be run.
       * @param loop_callback
       *  Callback to continue executing this function itself.
       */
      function run_notification(namespace, endpoint, pre_callback, post_callback, loop_callback, interval) {
        var ts_namespace = 'ts_' + namespace;
        var timeout = 15000; // ms

        // Check localstorage first
        var ts = parseInt(localStorage.getItem(ts_namespace));

        // See if our previous request timed out.
        if (localStorage.getItem('notification_running') && (Date.now() - localStorage.getItem('notification_running')) > timeout) {
          localStorage.removeItem('notification_running');
        }

        // Run update
        if ((isNaN(ts) || (Date.now() - ts) > (interval)) && !localStorage.getItem('notification_running')) {
          // console.log(namespace + ': running now @' + Date.now()/1000);
          // Set initial stage.
          pre_callback();

          // Set a semaphore
          localStorage.setItem('notification_running', Date.now());

          $.getJSON(endpoint, function(data) {
            var return_data = (data && data != -1) ? data : 'N/A';

            localStorage.setItem(namespace, JSON.stringify(return_data));
            localStorage.setItem(ts_namespace, Date.now());

            // Update UI once data gets returned.
            post_callback(return_data);

            // Recursive call for periodic updates.
            setTimeout(loop_callback, local_refresh_interval);

            // Remove semaphore
            localStorage.removeItem('notification_running');
          });
        }
        else {
          // console.log(namespace + ': ' + (Date.now() - ts)/1000 + 's elapsed');
          // Don't update, get update from localstorage
          var return_data = JSON.parse(localStorage.getItem(namespace));

          // Update UI once data.
          if (return_data) {
            post_callback(return_data);
          }

          // Recursive call for periodic updates.
          setTimeout(loop_callback, local_refresh_interval);
        }
      }

      /**
       * Run role based notification request.
       */
      function get_notification_badge() {
        run_notification('get_notification_badge', endpoint, function() {
          $badge_wrapper.find('.notifications-number')
            .addClass(loading_class)
            .text('loading...');
        }, function(return_data) {
          $badge_wrapper.find('.notifications-number')
            .removeClass(loading_class)
            .text(return_data);
        }, get_notification_badge, request_interval);
      }

      /**
       * Helper for formating user notification items.
       *
       * @param data
       * @returns {*}
       */
      function format_subscription_notifications(data) {
        if (typeof data != 'object') {
          return data;
        }

        var $ul = $('<ul />');

        $.each(data, function(k, item) {
          $ul.append('<li><span class="title">'+item.linked_title+'</span><span class="created">'+item.created+'</span></li>')
        });

        return $ul;
      }

      /**
       * Run user bsed notification request.
       */
      function get_subscriptions_badge() {
        run_notification('get_subscriptions_badge', subscriptions_endpoint, function() {
          $subscriptions_badge_wrapper.find('.notifications-wrapper')
            .addClass(loading_class)
            .text('loading...');
          $subscriptions_badge_wrapper.find('.notifications-number').addClass('loading').text('...');
        }, function(return_data) {
          $subscriptions_badge_wrapper.find('.notifications-number').removeClass('loading').text(return_data.length);
          $subscriptions_badge_wrapper.find('.notifications-wrapper')
            .removeClass(loading_class)
            .html(format_subscription_notifications(return_data));
        }, get_subscriptions_badge, subscriptions_request_interval);
      }
    }
  };

  /**
   * DataTables plugin applied on OpenEDU Notifications listing page.
   */
  Drupal.behaviors.OpenEDUNotificationsClientTable = {
    format: function(entity_id) {
      var group = Drupal.settings.openedu_notifications_client.nested_rows[entity_id];
      var $table = $('<table/>');
      $.each(group, function(k, v) {
        $('<tr/>')
          .append('<td>' + v.message + '</td>')
          .append('<td>' + v.entity_link + '</td>')
          .append('<td>' + v.created + '</td>')
          .appendTo($table);
      });

      return $table[0].outerHTML;
    },
    attach: function(context, settings) {
      // Set up vars.
      var that = this,
          notificationContainer = $('#openedu-notifications', context);

      // Check if plugin is available.
      if (notificationContainer.length && jQuery().DataTable) {
        // Look for table to be rendered.
        var notificationTable = $('#openedu-notifications-list').DataTable({
          "order": [[5, 'desc']]
        });

        // Add event listener for opening and closing details
        $('#openedu-notifications-list tbody').on('click', 'td.details-control a', function(e) {
          var tr = $(this).closest('tr'),
            row = notificationTable.row(tr);

          // This row is already open - close it
          if (row.child.isShown()) {
            row.child.hide();
            tr.removeClass('shown');
          }
          // Open this row
          else {
            row.child(that.format($(this).attr('rel'))).show();
            tr.addClass('shown');

            // equalize widths
            var left = tr.find('td:first-child').width() + tr.find('td:nth-child(2)').width() + 20*2;
            var right = tr.find('td:nth-child(6)').width() + 20;
            var width3 = tr.find('td:nth-child(3)').width();
            var width4 = tr.find('td:nth-child(4)').width();
            var width5 = tr.find('td:nth-child(5)').width();

            var $td = $(row.child()[0]).find('> td');
            $td.css({
              'padding-left': left,
              'padding-right': right
            });
            $td.find('td:nth-child(1)').width(width3);
            $td.find('td:nth-child(2)').width(width4);
            $td.find('td:nth-child(3)').width(width5);
          }
          e.preventDefault();
        });
      }
    }
  };

})(jQuery);
